/* Interrupt functions don't pass a pointer of the volume. So we have to include this
   information in the handle. */
int GetVolumeSpecificHandle(int Handle, int VolumeNr)
{
  if( (VolumeNr < MAX_DEVICE_COUNT) &&
      (devices[VolumeNr] != 0) ) {
    return VolumeNr*FAT_MAX_FILES+Handle+100; //first 100 handles are reserved for ipc-dos
  } else
    return -1;
}

// This function does the opposite of GetVolumeSpecificHandle. It returns
// the Volume-pointer and modifies the passed Handle.
// Returns 0 if handle belongs to an ipc-dos internal device
tFatVolume* GetVolumeByHandle(unsigned int *Handle)
{
  int OldHandle;
  if (*Handle < 100)
    return 0; // we access an internal device
  else {
    OldHandle = *Handle-100;
    *Handle = (*Handle-100) % FAT_MAX_FILES;
    return devices[OldHandle/FAT_MAX_FILES];
  }
}

// Function returns negative value, if we have to access an ipc-dos internal device
int GetVolumeByFilename(char** Filename)
{
  int Nr = defaultdevice;
  if (*Filename != 0) {
    if ((*Filename)[1] == ':') {
      if (*Filename[0] != 0) {
        Nr = toupper((*Filename)[0])-'C';
        if (Nr < 0)
          Nr = FD32_ERROR_INVALID_DRIVE;
        else if (Nr >= devicecount)
          Nr = FD32_ERROR_INVALID_DRIVE;
        else if (devices[Nr] == 0)
          Nr = FD32_ERROR_INVALID_DRIVE;
        (*Filename) += 2;
      }
    }
  }
  return Nr;
}

/* This Variable will be set true by int function drv_unmount*/
bool       driver_unmount = false;

/* OldHandler saves to old interrupt handler */
void interrupt (*OldHandler)(...);

#pragma argsused
/* This is the interrupt handler. See Documentation for Details */
void interrupt FatHandler(unsigned bp, unsigned di, unsigned si,
                          unsigned ds, unsigned es, unsigned dx,
                          unsigned cx, unsigned bx, unsigned ax,
                          unsigned ip, unsigned cs, unsigned flags)
{
  int           ret;
  long          LResult;
  int           IResult;
  char         *StringParam = (char*) MK_FP(es, di);
  int           DeviceNr;
  tFatVolume   *V;
  switch(ax) {
    case 0 :
    case 1 : //fat_mkdir
      ret = GetVolumeByFilename(&StringParam);
      if (ret >= 0)
        ret = fat_mkdir(devices[ret], StringParam, ax & 1);
      break;
    case 2 :
    case 3 : //fat_rmdir
      ret = GetVolumeByFilename(&StringParam);
      if (ret >= 0)
        ret = fat_rmdir(devices[ret], StringParam, ax & 1);
      break;
    case 4 :
    case 5 : //fat_open
      DeviceNr = GetVolumeByFilename(&StringParam);
      if (ret >= 0) {
        ret = fat_open(devices[DeviceNr], StringParam, (((unsigned long)dx) << 16) | cx, bx & 0xff, bx >> 8, &IResult, ax & 1);
        bx = IResult;
        if (ret >= 0)
          ret = GetVolumeSpecificHandle(ret, DeviceNr);
      }
      break;
    case 6 : //fat_close
      V = GetVolumeByHandle(&si);
      ret = fat_close(V, si);
      break;
    case 7 : //fat_fflush
      V = GetVolumeByHandle(&si);
      ret = fat_fflush(V, si);
      break;
    case 8 : //fat_read
      V = GetVolumeByHandle(&si);
      ret = fat_read(V, si, StringParam, (((unsigned long)dx) << 16) | cx, (unsigned long*) &LResult);
      dx = LResult >> 16;
      cx = LResult & 0xffff;
      break;
    case 9 : //fat_write
      V = GetVolumeByHandle(&si);
      ret = fat_write(V, si, StringParam, (((unsigned long)dx) << 16) | cx, (unsigned long*) &LResult);
      dx = LResult >> 16;
      cx = LResult & 0xffff;
      break;
    case 10 :
    case 11 : //fat_unlink
      ret = GetVolumeByFilename(&StringParam);
      if (ret >= 0)
        ret = fat_unlink(devices[ret], StringParam, bx, ax & 1);
      break;
    case 12 : //fat_lseek
      V = GetVolumeByHandle(&si);
      ret = fat_lseek(V, si, (((unsigned long)dx) << 16) | cx, bx, &LResult);
      dx = LResult >> 16;
      cx = LResult & 0xffff;
      break;
    case 13 : //fat_get_attr
      V = GetVolumeByHandle(&si);
      ret = fat_get_attr(V, si, (fd32_dev_fs_attr_t*) StringParam);
      bx = IResult;
      break;
    case 14 : //fat_set_attr
      V = GetVolumeByHandle(&si);
      ret = fat_set_attr(V, si, (fd32_dev_fs_attr_t*) StringParam);
      bx = IResult;
      break;
    case 15: //fat_dos_findfirst
      ret = GetVolumeByFilename(&StringParam);
      if (ret >= 0)
        ret = fat_dos_findfirst(ret, StringParam, bx, (fd32_fs_dosfind_t*) MK_FP(dx, cx));
      break;
    case 16: //fat_dos_findnext
      ret = fat_dos_findnext((fd32_fs_dosfind_t*) MK_FP(dx, cx));
      break;
    case 17: //fat_lfn_findfirst
      ret = GetVolumeByFilename(&StringParam);
      if (ret >= 0)
        ret = fat_lfn_findfirst(devices[ret], StringParam, bx,(fd32_fs_lfnfind_t*) MK_FP(dx, cx));
      break;
    case 18: //fat_lfn_findnext
      ret = fat_lfn_findnext((fd32_fs_lfnfind_t*) MK_FP(dx, cx));
      break;
    case 19: //fat_lfn_findclose
      ret = fat_lfn_findclose((fd32_fs_lfnfind_t*) MK_FP(dx, cx));
      break;
    case 20:
    case 21: //fat_rename
      ret = GetVolumeByFilename(&StringParam);
      if (ret >= 0)
        ret = fat_rename(devices[ret], StringParam, (char*) MK_FP(dx, cx), ax & 1);
      break;
    case 22: //drv_unmount
      driver_unmount = true;
      break;
    case 23: //fat_fflush
      IResult = si;
      V = GetVolumeByHandle((unsigned*) &IResult);
      if (V != 0)
        ret = fat_fflush(V, IResult);
      break;
    case 24: //set default drive
      if ((bx & 0xff) < devicecount) {
        defaultdevice = bx & 0xff;
        ret = devicecount;
      } else
        ret = FD32_ERROR_INVALID_DRIVE;
      break;
    case 25:  //get default drive
      ret = defaultdevice;
      break;
    default:
      ret = 0;
      break;
  }
  ax = ret;
}

/*
void interrupt Int21Handler(unsigned bp, unsigned di, unsigned si,
                            unsigned ds, unsigned es, unsigned dx,
                            unsigned cx, unsigned bx, unsigned ax,
                            unsigned ip, unsigned cs, unsigned flags)
{
  int           ret;

  unsigned      UResult1, UResult2;
  long          LResult;
  unsigned long ULResult;
  int           IResult;
  char         *StringParam;
  int           DeviceNr;
  tFatVolume   *V;
  bool          isInternal=true;

  StringParam = (char*) MK_FP(ds, dx);
  isInternal = true;

  switch(ax >> 8) {
    case 0x0e: //DOS 1+ - SELECT DEFAULT DRIVE
printf("set drive\n");
      defaultdevice = (dx & 0xff)-2;
      if (defaultdevice >= 2) {
        isInternal = false;
        ax = 32;
      }
      break;
    case 0x19: //DOS 1+ - GET CURRENT DEFAULT DRIVE
printf("get drive\n");
      ax = defaultdevice+2;
      isInternal = false;
      break;
    case 0x36: //DOS 2+ - GET FREE DISK SPACE
printf("get diskspace\n");
      IResult = ((int)dx & 0xff)-2;
      if ((IResult >= 0) && (IResult < MAX_DEVICE_COUNT)) {
        V = devices[IResult];
        ax = V->Bpb.BPB_SecPerClus;
        bx = V->FSI_Free_Count;
        cx = V->Bpb.BPB_BytsPerSec;
        dx = V->DataClusters;
        isInternal=false;
      }
      break;
    case 0x39: //DOS 2+ - MKDIR - CREATE SUBDIRECTORY
printf("create directory\n");
      IResult = GetVolumeByFilename(&StringParam);
      if (IResult >= 0) {
        V = devices[IResult];
        ret = fat_mkdir(V, StringParam, 0);
        if (ret != 0) {
          flags |= 0x0001;
          ax = -ret;
        } else
          flags &= 0xfffe;
        isInternal=false;
      }
      break;
    case 0x3a: //DOS 2+ - RMDIR - REMOVE SUBDIRECTORY
printf("rmdir\n");
      IResult = GetVolumeByFilename(&StringParam);
      if (IResult >= 0) {
        V = devices[IResult];
        ret = fat_mkdir(V, StringParam, 0);
        if (ret != 0) {
          flags |= 0x0001;
          ax = -ret;
        } else
          flags &= 0xfffe;
        isInternal=false;
      }
      break;
//chdir
    case 0x3c: //DOS 2+ - CREAT - CREATE OR TRUNCATE FILE
printf("create file\n");
      IResult = GetVolumeByFilename(&StringParam);
      if (IResult >= 0) {
        V = devices[IResult];
        ret = fat_open(V, StringParam, FD32_OPEN_EXIST_REPLACE |
                                       FD32_OPEN_NOTEXIST_CREATE |
                                       FD32_OPEN_ACCESS_READWRITE |
                                       FD32_OPEN_SHARE_DENYALL, cx , 1, (int*) &UResult1, 0);
        if (ret < 0) {
          flags |= 0x0001;
          ax = -ret;
        } else {
          flags &= 0xfffe;
          ax = GetVolumeSpecificHandle(ret, IResult);
        }
        isInternal=false;
      }
      break;
    case 0x3d: //DOS 2+ - OPEN - OPEN EXISTING FILE
printf("open file\n");
      IResult = GetVolumeByFilename(&StringParam);
      if (IResult >= 0) {
        V = devices[IResult];
        ret = fat_open(V, StringParam, FD32_OPEN_EXIST_OPEN | (ax & 0xff), 0xff, 1, (int*) &UResult1, 0);
        if (ret < 0) {
          flags |= 0x0001;
          ax = -ret;
        } else {
          flags &= 0xfffe;
          ax = GetVolumeSpecificHandle(ret, IResult);
printf("opened handle %u\n", ax);
        }
        isInternal=false;
      }
      break;
    case 0x3e: //DOS 2+ - CLOSE - CLOSE FILE
printf("close file handle %d\n", bx);
      IResult = bx;
      V = GetVolumeByHandle((unsigned*) &IResult);
      if (V != 0) { //Check, if we access an ipc-dos internal handle
        ret = fat_close(V, IResult);
        if (ret != 0) {
          flags |= 0x0001;
          ax = -ret;
        } else {
          flags &= 0xfffe;
        }
        isInternal = false;
      }
      break;
    case 0x3f: //DOS 2+ - READ - READ FROM FILE OR DEVICE
printf("Handle=%u\n", bx);
printf("reading %ud bytes\n", cx);
      UResult1 = bx;
      V = GetVolumeByHandle(&UResult1);
      if (V != 0) { //Check, if we access an ipc-dos internal handle
        ret = fat_read(V, UResult1, StringParam, cx, &ULResult);
        if (ret != 0) {
          flags |= 0x0001;
          ax = -ret;
        } else {
          ax = (unsigned) ULResult; //Number of bytes read
          flags &= 0xfffe;
        }
        isInternal = false;
      }
      break;
    case 0x40: //DOS 2+ - WRITE - WRITE TO FILE OR DEVICE
      UResult1 = bx;
      V = GetVolumeByHandle(&UResult1);
      if (V != 0) { //Check, if we access an ipc-dos internal handle
        ret = fat_write(V, UResult1, StringParam, cx, &ULResult);
        if (ret != 0) {
          flags |= 0x0001;
          ax = -ret;
        } else {
          ax = (unsigned) ULResult; //Number of bytes written
          flags &= 0xfffe;
        }
        isInternal = false;
      }
      break;
    case 0x41: //DOS 2+ - UNLINK - DELETE FILE
printf("unlink\n");
      IResult = GetVolumeByFilename(&StringParam);
      if (IResult >= 0) {
        V = devices[IResult];
        ret = fat_unlink(V, StringParam, (unsigned long) (cx & 0xff), 0);
        if (ret != 0) {
          flags |= 0x0001;
          ax = -ret;
        } else {
          flags &= 0xfffe;
        }
        isInternal=false;
      }
      break;
    case 0x42: //DOS 2+ - LSEEK - SET CURRENT FILE POSITION
printf("lseek\n");
      IResult = bx;
      V = GetVolumeByHandle((unsigned*) &IResult);
      if (V != 0) { //Check, if we access an ipc-dos internal handle
        ret = fat_lseek(V, IResult, (((long) cx) << 16) | ((long) dx), ax & 0xff, &LResult);
        if (ret != 0) {
          flags |= 0x0001;
          ax = -ret;
        } else {
          dx = LResult >> 16;
          ax = LResult & 0xffff; //=New Position relative to start of file
          flags &= 0xfffe;
        }
        isInternal = false;
      }
      break;
//*********** Luft bei DOS ber Dateiname!!!!
    case 0x43: //DOS 2+ - GET/SET FILE ATTRIBUTES
printf("file attribs\n");
//      if (ax & 0xff) { //al.0 = 1 => SET
//        ret = fat_set_attr(V, cx, (fd32_dev_fs_attr_t*) StringParam);
//      } else {         //al.0 = 0 => GET
//        ret = fat_get_attr(V, cx, (fd32_dev_fs_attr_t*) StringParam);
//      }
      cx = 1+32;
isInternal = false;      
      flags &= 0xfffe;
      break;
    case 0x44: //DOS 2+ - IOCTL Set/Get Device Information
printf("ioctl handle=%u\n", bx);
printf("ioctl subfunction=%u\n", ax & 0xff);
      UResult1 = bx;
      V = GetVolumeByHandle(&UResult1);
      if (V != 0) { //Check, if we access an ipc-dos internal handle
        if( (ax & 0x0001) == 0 ) {  //Get Device Data
          flags &= 0xfffe;
          dx =  0x40 |(2+(bx-100)/FAT_MAX_FILES); //return Drive number. Bit 6: Set to '0' when the file has been written
printf("get ioctl dx=%x\n", dx);
        } else {                    //Set Device Data
          flags &= 0xfffe;
printf("set ioctl\n");
        }
        isInternal = false;
      }
      break;
//CWD
    case 0x4e: //DOS 2+ - FINDFIRST - FIND FIRST MATCHING FILE
printf("findfirst\n");
      IResult = GetVolumeByFilename(&StringParam);
      asm {
        mov ah,0x2f
        int 0x30
        mov UResult1, es
        mov UResult2, bx
      }
      if( (IResult >= 0) && (IResult < devicecount) ) {
        ret = fat_dos_findfirst(IResult, StringParam, (cx & 0xff), (fd32_fs_dosfind_t*) MK_FP(UResult1, UResult2));
        if (ret < 0) {
          flags |= 0x0001;
          ax = -ret;
        } else {
          flags &= 0xfffe;
        }
        isInternal=false;
      } else {
        //ipc-dos does not make ANY use of reserved searchstructure fields,
        //so we are able to use it, to mark it as ipc-dos internal structure.
        //This enables findnext to detect an ipc-dos internal structure.
        ((tDosFind*) MK_FP(UResult1, UResult2))->Drive = 0xff;
      }
      break;
    case 0x4f: //DOS 2+ - FINDNEXT - FIND NEXT MATCHING FILE
printf("findnext\n");
      asm {
        mov ah,0x2f
        int 0x30
        mov UResult1, es
        mov UResult2, bx
      }
      StringParam = (char*) MK_FP(UResult1, UResult2);
      //Check, if this is an fd32 searchstructure...
      if (((tDosFind*) StringParam)->Drive != 0xff ) {
        ret = fat_dos_findnext((fd32_fs_dosfind_t*) StringParam);
        if (ret < 0) {
          flags |= 0x0001;
          ax = -ret;
        } else
          flags &= 0xfffe;
        isInternal = false;
      }
      break;
    case 0x56: //DOS 2+ - RENAME - RENAME FILE
printf("rename\n");
      IResult = GetVolumeByFilename(&StringParam);
      if (IResult >= 0) {
        V = devices[IResult];
        ret = fat_rename(V, StringParam, (char*) MK_FP(es, di), 0);
        isInternal=false;
      }
      break;
// GET/SET FILES LAST WRITTEN DATE AND TIME
    case 0x68: //DOS 3.3+ - FFLUSH - COMMIT FILE
printf("fflush\n");
      IResult = bx;
      V = GetVolumeByHandle((unsigned*) &IResult);
      if (V != 0) { //Check, if we access an ipc-dos internal handle
        ret = fat_fflush(V, IResult);
        if (ret != 0) {
          flags |= 0x0001;
          ax = -ret;
        } else {
          flags &= 0xfffe;
        }
        isInternal = false;
      }
      break;
    default:
      printf("ax=%x\n", ax);
      isInternal = true;
      break;
  }
  if (isInternal) {
    struct REGPACK reg;
    reg.r_ax = ax; reg.r_bx = bx; reg.r_cx = cx; reg.r_dx = dx;
    reg.r_bp = bp; reg.r_si = si; reg.r_di = di; reg.r_ds = ds;
    reg.r_es = es; reg.r_flags = flags;
    intr(0x30, &reg);
if( (ax >> 8) == 0x44)
  printf("get ioctlINTERN dx=%x\n", reg.r_dx);

    ax = reg.r_ax; bx = reg.r_bx; cx = reg.r_cx; dx = reg.r_dx;
    bp = reg.r_bp; si = reg.r_si; di = reg.r_di; ds = reg.r_ds;
    es = reg.r_es; flags = reg.r_flags;
  }
}*/

void InstallIntHandler(void)
{
  OldHandler = getvect(0x30);
  *(unsigned*)(0x30*4 + 0) = FP_OFF(FatHandler);
  *(unsigned*)(0x30*4 + 2) = FP_SEG(FatHandler);
  printf("Installed Interrupt handler\n");
}

void UnInstallIntHandler(void)
{
  *(unsigned*)(0x30*4 + 0) = FP_OFF(OldHandler);
  *(unsigned*)(0x30*4 + 2) = FP_SEG(OldHandler);
  printf("Uninstalled Interrupt handler\n");
}


